/* Styles for main page */
#banner {
  background: #9c9;
  padding-top: 5px;
  padding-bottom: 5px;
  border-bottom: 2px solid;
  font: small-caps 20px/20px "Times New Roman", serif;
  color: #282;
  text-align: center;
}

#banner img {
  float: left;
}

#main {
  margin-left: 0em;
  padding-top: 4ex;
  padding-left: 2em;
  background: white;
}

h1 {
  font:  150% sans-serif;
  color: #226;
  border-bottom: 3px dotted #77d;
}

body {
    font: normal 75% verdana,arial,helvetica;
    color:#000000;
}

table tr td, table tr th {
  font-size: 75%;
}

table.stripes tr th {
    font-weight: bold;
    text-align: left;
    background: #a0a0a0;
}

table.stripes tr td {
    background: #ccccc0;
}

td.pass {
  color: green;
}

td.fail {
  color: red;
  font-weight: bold;
}

#main {
    padding-left: 0em;
}

#controls {
  float: left;
  padding-top: 1em;
  padding-left: 1em;
  padding-right: 1em;
  padding-bottom: 1em;
  width: 14em;
  border-right: 2px solid;
}

#body {
  margin-left: 16em;
  padding-top: 4ex;
  padding-left: 2em;
  background: white;
  border-left: 2px solid;
}
